#!/usr/bin/perl -Tw
################################################################################
package WWW::Yahoo::Briefcase;
# $Id: Briefcase.pm,v 1.4 2003/12/05 00:03:19 ken Exp $

=pod

=head1 NAME

WWW::Yahoo::Briefcase - interface to Yahoo Briefcase services

=head1 SYNOPSIS

  require WWW::Yahoo::Briefcase;

  my $cookies =
    WWW::Yahoo::Credentials::File->new( subdir => '.yahoo',
					filename => 'cookies.txt' );
  my $ybc = WWW::Yahoo::Briefcase->new( cookie_jar => $cookies,
					overwrite => 1 );
  $ybc->signin( $id, $pw );
  $ybc->create( '/', 'My Backups' ); # create 'My Backups' folder in '/'
  $ybc->upload( 'My Backups', '/var/backups/dump.bz2' );
  $ybc->upload( 'My Photos', 'picture.jpg' );
  $ybc->rename( 'My Backups', 'dump.bz2' => 'backup.bz2' );
  $ybc->move( 'My Photos' => 'My Backups', 'picture.jpg' );
  $ybc->copy( 'My Photos' => 'My Backups', 'picture.jpg' );
  $ybc->delete( 'My Photos', 'picture.jpg' );
  $ybc->delete( '/', 'My Photos' ); # delete all photos
  my @backups = $ybc->list( 'My Backups' );

=head1 DESCRIPTION

This module provides access to your Yahoo Briefcase.  You can upload
and download files and perform other operations such as list, rename,
move, copy, delete, and create new folders.

Errors may be caught using "eval", for example,

See
L<WWW::Yahoo::Briefcase::Exceptions|WWW::Yahoo::Briefcase::Exceptions>
for a list of exceptions (they are thrown using the
L<Exception::Class|Exception::Class> module).

  eval { $ybc->upload( $folder, @objects ) };
  if ( $@ ) {
      if ( $@->isa( 'WWW::Yahoo::Briefcase::Exception::List' ) ) {
	  my @errors = map { ucfirst( $_->description ) . $_ } $@->errors;
	  die join( "\n", @errors ) . "\n";
      }
      else {
	  die ucfirst( $@->description ) . ": $@";
      }
  }

=head1 METHODS

=cut

################################################################################
use 5.006;
use strict;
use warnings;
use base qw( WWW::Mechanize );
require URI;
require HTML::TableExtract;
require Time::Local;
require WWW::Yahoo::Briefcase::Object;
require WWW::Yahoo::Briefcase::Exceptions;

our $VERSION = '0.81';
our $upload_attempts = 3; # number of times to attempt an upload
our $list_cache_time = 30; # seconds to cache a folder listing
my $s = '(?:[\s\xA0]|&nbsp;?|&\#160;?)'; # various whitespace characters

################################################################################

=pod

=head2 new

  my $ybc = new WWW::Yahoo::Briefcase;
  my $ybc = WWW::Yahoo::Briefcase->new( cookie_jar => $cookies,
					overwrite => 1, sleep => 3,
					debug => 0, maxsize => 10_485_760 );

Creates a new object.  Options may be specified when the object is
created.  The following options have default values:

  cookie_jar: {}
  overwrite:  undef
  sleep:      3 seconds
  debug:      off
  maxsize:    5,242,880 bytes

The cookie jar can be an already-initialized HTTP::Cookies object, for
example:
  my $cookies = HTTP::Cookies->new( file => $cookie_file, autosave => 1 );
Alternatively, the cookie file can be specified in which case a new
HTTP::Cookies object will be initialized from that file, for example:
  my $ybc = WWW::Yahoo::Briefcase->new( cookie_jar => $cookie_file );
If the file does not exist, it will be created with the umask set to
077 to protect it (the umask will be reset to its original value after
the cookie file is created).  If no cookie jar is specified, then an
in-memory cookie jar will be used ({}).

Any other options specified are passed to the superclasses,
L<WWW::Mechanize|WWW::Mechanize> and L<LWP::UserAgent|LWP::UserAgent>,
see their documentation for more information.

=cut

#-------------------------------------------------------------------------------
sub new {
    my $class = shift;
    my %options = ( overwrite => undef, sleep => 3, debug => 0,
		    maxsize => 5_242_880, @_ );

    # Initialize cookie jar (if needed).
    $options{cookie_jar} = {} unless defined $options{cookie_jar};
    unless ( UNIVERSAL::isa( $options{cookie_jar}, 'HTTP::Cookies' )
	     or ref( $options{cookie_jar} ) eq 'HASH' )
    {
	my $umask = umask 077; # to protect cookie file
	$options{cookie_jar} =~ m!(.*)!; # to untaint
	$options{cookie_jar} = HTTP::Cookies->new( file => $1, autosave => 1 );
	umask $umask;
    }

    # Remove our options before passing the rest up to the superclass.
    my @ours = qw( overwrite sleep debug maxsize );
    my( $overwrite, $sleep, $debug, $maxsize ) = @options{@ours};
    delete @options{@ours};

    my $self = $class->SUPER::new( %options );

    $self->overwrite( $overwrite );
    $self->sleep( $sleep );
    $self->debug( $debug );
    $self->maxsize( $maxsize );

    $self;
}

################################################################################

=pod

=head2 overwrite

  my $old = $ybc->overwrite( undef ); # do not overwrite
  my $old = $ybc->overwrite( '' ); # do not overwrite
  my $old = $ybc->overwrite( 0 ); # do not overwrite
  my $old = $ybc->overwrite( 1 ); # overwrite (pre-delete)
  my $old = $ybc->overwrite( 'NEW' ); # overwrite with prefix (post-delete)
  my $overwrite = $ybc->overwrite;

If given a parameter, it sets the overwrite flag.  With no parameter
specified the current setting of the prefix is returned.  If the
overwrite prefix is true, then existing objects will be overwritten on
upload, download, create, rename, move, or copy.

If the overwrite flag is set to 1, then existing objects with the same
name will be deleted I<before> the operation is preformed.  For
example, if the file 'backup.tar.bz2' is being uploaded and there
already existed such a file in the destination folder, then that file
would be deleted before 'backup.tar.bz2' is uploaded.

One caveat with this approach is that if the upload fails, then you've
lost the previous backup.  So, to avoid this problem, you can set the
overwrite flag to any true value other than 1 and it will be used as a
prefix for the filename.  If set to 'NEW', for example, then the file
in the above example would be uploaded as 'NEW-backup.tar.bz2', and
only if the upload succeeded would the original file 'backup.tar.bz2'
be deleted and 'NEW-backup.tar.bz2 renamed to 'backup.tar.bz2'.  One
caveat with this approach is that it takes up twice the space while
the new file(s) are being uploaded (all files are uploaded before any
files are deleted).

If you would like to specify a prefix of 0 or 1 for post-delete, then
specify the full prefix including the dash as '0-' or '1-'.

=cut

#-------------------------------------------------------------------------------
sub overwrite {
    my $self = shift;
    my $old = $self->{__PACKAGE__.'_overwrite'};
    if ( @_ ) {
	if ( not $_[0] ) {
	    $self->{__PACKAGE__.'_overwrite'} = undef;
	}
	elsif ( $_[0] eq '1' ) {
	    $self->{__PACKAGE__.'_overwrite'} = '1';
	}
	elsif ( substr( $_[0], -1 ) eq '-' ) {
	    $self->{__PACKAGE__.'_overwrite'} = $_[0];
	}
	else {
	    $self->{__PACKAGE__.'_overwrite'} = "$_[0]-";
	}
    }
    $old;
}

################################################################################

=pod

=head2 sleep

  my $old = $ybc->sleep( 3 );
  my $sleep = $ybc->sleep;

If given a parameter, it sets the numbers of seconds to sleep.
Otherwise, it returns the number.  If C<sleep> is set, then
C<get>, C<click>, and C<submit> (and anything else that calls LWP's
C<request> method which we've overridden) will C<sleep> for the
specified period before each fetch.  I ran into problems when setting
sleep to zero--only about half of the files were uploaded
successfully.

=cut

#-------------------------------------------------------------------------------
sub sleep {
    my $self = shift;
    my $old = $self->{__PACKAGE__.'_sleep'};
    $self->{__PACKAGE__.'_sleep'} = shift if @_;
    $old;
}

#-------------------------------------------------------------------------------
sub send_request {
    my $self = shift;
    my $request = $_[0];
    my $response = $self->SUPER::send_request( @_ );
    $self->_debug_rr( $request, $response ) if $self->debug > 1;
    $response;
}

#-------------------------------------------------------------------------------
sub request {
    my $self = shift;
    delete $self->{__PACKAGE__.'_list_cache'}; # invalidate the list cache
    my $requests_sent = $self->{__PACKAGE__.'_requests_sent'};
    CORE::sleep( $self->sleep ) if $self->sleep and $requests_sent;
    ++$self->{__PACKAGE__.'_requests_sent'};
    my $response = $self->SUPER::request( @_ );
    unless ( $response->is_success ) {
	my $error = "request failed";
	my @args = ( error => $error, request => $_[0], response => $response );
	WWW::Yahoo::Briefcase::Exception::Server->throw( @args );
    }
    $response;
}

#-------------------------------------------------------------------------------
# TODO: remove this subroutine when WWW::Mechanize is updated with patch I sent.
sub redirect_ok {
    my $self = shift;
    my $prospective_request = shift;

    my $method = $prospective_request->method;
    my $ok = $self->SUPER::redirect_ok( $prospective_request );
    if ( $ok and $method eq 'POST' ) {
	# Mimic erroneous browser behavior by changing the method and
	# removing content.
	$prospective_request->method( 'GET' );
	$prospective_request->content( '' );
	$prospective_request->remove_header( 'Content-Length' );
    }
    $ok;
};

################################################################################

=pod

=head2 debug

  my $old = $ybc->debug( 0 ); # turn off debugging
  my $old = $ybc->debug( 1 ); # show informational messages
  my $old = $ybc->debug( 2 ); # also show requests and responses
  my $debug = $ybc->debug;

Sets debugging level.  When enabled, information is printed to STDERR.

=cut

#-------------------------------------------------------------------------------
sub debug {
    my $self = shift;
    my $old = $self->{__PACKAGE__.'_debug'};
    $self->{__PACKAGE__.'_debug'} = shift if @_;
    $old;
}

#-------------------------------------------------------------------------------
sub _debug_rr {
    my $self = shift;
    my( $request, $response ) = @_;

    if ( $request ) {
	warn "-- REQUEST ---------------------------------------------------\n";
	warn $request->method . ' ' . $request->uri . "\n";
	warn $request->headers_as_string . "\n";
	warn $request->content . "\n";
    }
    else {
	warn "-- NO REQUEST ------------------------------------------------\n";
    }

    if ( $response ) {
	warn "-- RESPONSE --------------------------------------------------\n";
	my $protocol = $response->protocol;
	$protocol = "(Protocol Undefined)" unless defined $protocol;
	warn $protocol . ' ' . $response->status_line . "\n";
	$response->scan( sub {
			     return if $_[0] =~ m!^Client-!; # ignore LWP header
			     warn "$_[0]: $_[1]\n";
			 } );
	warn "\n";
	my $content = $response->content;
	$content .= "\n" unless substr( $content, -1 ) eq "\n";
	warn $content;
    }
    else {
	warn "-- NO RESPONSE -----------------------------------------------\n";
    }

    warn "--------------------------------------------------------------\n";
}

################################################################################

=pod

=head2 maxsize

  my $old = $ybc->maxsize( 10_485_760 );
  my $maxsize = $ybc->maxsize;

If given a parameter, it sets the maximum size of a file upload.
Otherwise, it returns the size.  Note that this setting is not a user
preference but is instead a limitation imposed on uploads to your
briefcase.  That is, you have no control over it.  This method exists
solely so that you can inform this class what your limit is, so that
the class will not attempt to upload files larger than the limit.

The standard file size limit is 5MB, but with Premium Service it is
10MB.

=cut

#-------------------------------------------------------------------------------
sub maxsize {
    my $self = shift;
    my $old = $self->{__PACKAGE__.'_maxsize'};
    $self->{__PACKAGE__.'_maxsize'} = shift if @_;
    $old;
}

################################################################################

=pod

=head2 do

  my @commands = $WWW::Yahoo::Briefcase->do;
  $ybc->do( $command, @arguments );
  $ybc->do( 'signin', $id, $pw );
  $ybc->do( 'signout' );
  $ybc->do( 'upload', $folder, @files );
  $ybc->do( 'download', $folder, @files );
  $ybc->do( 'create', $folder, @subfolders );
  $ybc->do( 'rename', $folder, $object => $newname );
  $ybc->do( 'delete', $folder, @objects );
  $ybc->do( 'move', $folder => $destination, @objects );
  $ybc->do( 'copy', $folder => $destination, @objects );
  my @objects = $ybc->do( 'list', $folder );

Return list of supported commands or perform specified command.

=cut

#-------------------------------------------------------------------------------
sub do {
    my $self = shift;

    my @commands =
      qw( signin signout upload download create rename delete move copy list );
    return @commands unless @_;

    my $command = shift;
    unless ( grep { $command eq $_ } @commands ) {
	my $error = "invalid command: \"$command\"";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }

    no strict 'refs';
    $self->$command( @_ );
}

################################################################################

=pod

=head2 signin

  $ybc->signin( $id, $pw );

Sign in if not already signed in.  Gets/updates cookie and grabs a few
links that will be needed.

=cut

#-------------------------------------------------------------------------------
sub signin {
    my $self = shift;
    my( $id, $pw ) = ( shift, shift );
    if ( @_ ) {
	my $error = "extra arguments specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    if ( not defined $id or $id eq '' ) {
	WWW::Yahoo::Briefcase::Exception::Usage->throw( "id not specified" );
    }
    if ( not defined $pw or $pw eq '' ) {
	WWW::Yahoo::Briefcase::Exception::Usage->throw( "pw not specified" );
    }

    # Get main page.
    $self->get( 'http://briefcase.yahoo.com/' );

    if ( $self->content =~ m!Sign${s}In|confirm${s}your${s}password! ) {
	# Fill out and submit form to sign in.
	warn "INFO: Signing in\n" if $self->debug;
	$self->_signin_form( $id, $pw );
    }
    if ( $self->content =~ m!Sign${s}In|confirm${s}your${s}password! ) {
	my $error = "unsuccessful (sign in/confirm still required)";
	WWW::Yahoo::Briefcase::Exception::Signin->throw( $error );
    }

    # Extract links that will be useful later.
    $self->_signin_links;

    (); # no return value
}

#-------------------------------------------------------------------------------
sub _signin_form {
    my $self = shift;
    my( $id, $pw ) = @_;

    # Note: this subroutine handles both initial sign-in and "Verify
    # your Yahoo! password to continue".  In the latter case, there is
    # no 'login' field nor '.persistent' field.

    my $form = $self->current_form;
    unless ( $form->action =~ m!^https:! ) {
	my $error = "URL for sign form is not secure: " . $form->action;
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }

    eval {
	$self->field( login => $id ) if $form->find_input( 'login' );
	$self->field( passwd => $pw );
	$self->field( '.persistent' => 'Y' )
	  if $form->find_input( '.persistent' );
	$self->click;
	if ( $self->content =~ m!window\.location\.replace\("([^\"]*)"\)! ) {
	    $self->get( $1 );
	}
    };
    WWW::Yahoo::Briefcase::Exception::Interface->throw( $@ ) if $@;
}

#-------------------------------------------------------------------------------
sub _signin_links {
    my $self = shift;

    # Get link for signing out.
    my $link_signout = $self->find_link( text_regex => qr!Sign${s}Out! );
    unless ( defined $link_signout ) {
	my $error = "no link to sign out";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
    $link_signout = URI->new_abs( $link_signout->url, $self->base );
    $self->{__PACKAGE__.'_link_signout'} = $link_signout;

    # Get link for listing folders.
    my $link_myfolders = $self->find_link( text_regex => qr!My${s}Folders! );
    unless ( defined $link_myfolders ) {
	my $error = "no link to list folders";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
    $link_myfolders = URI->new_abs( $link_myfolders->url, $self->base );
    $self->{__PACKAGE__.'_link_myfolders'} = $link_myfolders;

    # Get link for form to select destination folder for adding files.
    my $link_addfiles = $self->find_link( text_regex => qr!Add${s}Files! );
    unless ( defined $link_addfiles ) {
	my $error = "no link to add files";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
    $link_addfiles = URI->new_abs( $link_addfiles->url, $self->base );
    $self->{__PACKAGE__.'_link_addfiles'} = $link_addfiles;
}

################################################################################

=pod

=head2 signout

  $ybc->signout;

Sign out.  If you are using a cookie file, then do not call this
function unless you want to invalidate your cookies.  The cookie file
allows you to access your briefcase more quickly because the login
form is bypassed (assuming your cookie has not timed out).

=cut

#-------------------------------------------------------------------------------
sub signout {
    my $self = shift;
    warn "INFO: Signing out\n" if $self->debug;
    if ( @_ ) {
	my $error = "extra arguments specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }

    # Request sign out.
    $self->get( $self->{__PACKAGE__.'_link_signout'} );
    if ( $self->content =~ m!Sign${s}Out! ) {
	my $error = "unexpected response for sign out";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }

    # Get main page to verify that we have signed out.
    $self->get( 'http://briefcase.yahoo.com/' );
    unless ( $self->content =~ m!Sign${s}In! ) {
	my $error = "unexpected response for main page";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }

    (); # no return value
}

################################################################################

=pod

=head2 upload

  $ybc->upload( $folder, @files );
  $ybc->upload( $folder, $file );
  $ybc->upload( $folder, [ $file1 => $newfilename1 ], $file2, ... );

  $ybc->upload( 'My Backups', '/var/backups/dump.bz2' );
  $ybc->upload( 'My Backups', [ '/var/backups/dump.bz2' => '20030415.bz2' ] );

Upload file(s) to specified folder.  The full path name of the
existing folder on the remote system must be specified, and the full
path name of each existing file you wish to upload from your local
system must be specified (or the relative path name of an existing
file if it is reachable from the current directory).  You can give
each file a new name, but be sure to include the extension.

Note that if an existing object of the same name is found in the
destination folder, it will automatically be renamed (for example,
'pic.jpg' might become 'pic---1.jpg'), unless you have enabled the
overwrite flag, in which case the object will be overwritten.  Note
well that if the pre-existing object to be overwritten is a folder
then all its contents are deleted!

=cut

################################################################################
sub upload {
    my $self = shift;
    if ( @_ < 2 ) {
	my $error = "folder and/or file(s) not specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    my $folder = $self->_canonical_folder( shift );

    # Simply upload files unless we need to deal with overwriting files.
    return $self->_upload_files( $folder, @_ ) unless $self->overwrite;

    # Split each filename into [ $path, $name, $type ].
    my @files;
    foreach my $file ( @_ ) {
	my( $path, $type ) = ( '', '' );
	my $name = $file;
	$name =~ s!^(.*/)!! and $path = $1;
	$name =~ s!(\.[^.]*)$!! and $type = $1;
	push( @files, [ $path, $name, $type ] );
    }

    # Delete old files with prefix from folder.
    my $prefix = $self->overwrite;
    my $exists = $self->_delete_if_exists( $folder, $prefix, \@files );
    $prefix = '' if $prefix eq '1'; # pre-delete files

    # Upload new files to folder.
    my( @upload, $need_to_rename );
    foreach my $file ( @files ) {
	my( $path, $name, $type ) = @$file;
	if ( $prefix and $exists->{"$name$type"} ) {
	    ++$need_to_rename;
	    push( @upload, [ "$path$name$type" => "$prefix$name$type" ] );
	}
	else {
	    push( @upload, "$path$name$type" );
	}
    }
    eval { $self->_upload_files( $folder, @upload ) };
    unless ( $need_to_rename ) {
	$@->rethrow if $@;
	return;
    }
    my @errors = $@ ? ( $@ ) : ();

    # Delete old files from folder.
    eval { $exists = $self->_delete_if_exists( $folder, '', \@files ) };
    WWW::Yahoo::Briefcase::Exception::List->throw_list( @errors, $@ ) if $@;

    # Remove prefix from new files.
    my $overwrite = $self->overwrite( 0 ); # inhibit extraneous listings
    foreach my $file ( @files ) {
	my( $path, $name, $type ) = @$file;
	if ( $prefix and $exists->{"$prefix$name$type"} ) {
	    eval { $self->rename( $folder, "$prefix$name$type" => "$name" ) };
	    push( @errors, $@ ) if $@;
	}
    }
    $self->overwrite( $overwrite ); # restore previous value

    WWW::Yahoo::Briefcase::Exception::List->throw_list( @errors ) if @errors;
    (); # no return value
}

#-------------------------------------------------------------------------------
sub _delete_if_exists {
    my $self = shift;
    my( $folder, $which, $files ) = @_;
    my %exists = map { $_ => 1 } $self->list( $folder );
    my @delete;
    my $prefix = $self->overwrite;
    foreach my $file ( @$files ) {
	my( $path, $name, $type ) = @$file;
	if ( $prefix eq '1' ) {
	    # Pre-delete files.
	    push( @delete, "$name$type" ) if $exists{"$name$type"};
	}
	elsif ( $exists{"$prefix$name$type"} ) {
	    # Post-delete files.
	    push( @delete, "$which$name$type" ) if $exists{"$which$name$type"};
	}
    }
    $self->delete( $folder, @delete ) if @delete;
    \%exists;
}

#-------------------------------------------------------------------------------
sub _upload_files {
    my $self = shift;
    my $folder = shift;
    my @errors;
    foreach my $file ( @_ ) {
	my $filename = ref $file ? $file->[1] : undef;
	$file = $file->[0] if ref $file;
	my $attempts = $upload_attempts;
	while ( $attempts > 0 ) {
	    eval { $self->_upload_file( $folder, $file, $filename ) };
	    last unless $@;
	    last unless $@->isa( 'WWW::Yahoo::Briefcase::Exception::Server' );
	    my $status = $@->response->status_line;
	    last if $status eq "500 Can't connect";
	    CORE::sleep 5;
	    --$attempts
	      and warn "WARN: upload failed ($status); trying again: $file\n";
	}
	push( @errors, $@ ) if $@;
    }
    WWW::Yahoo::Briefcase::Exception::List->throw_list( @errors ) if @errors;
}

#-------------------------------------------------------------------------------
sub _upload_file {
    my $self = shift;
    my( $folder, $file, $filename ) = @_;

    if ( $self->debug ) {
	my $as = defined $filename ? " as \"$filename\"" : '';
	warn "INFO: Uploading \"$file\" to \"$folder\"$as\n";
    }

    # Check arguments.
    unless ( defined $filename ) {
	( $filename ) = $file =~ m!([^/]*)$!;
    }
    -e $file
      or WWW::Yahoo::Briefcase::Exception::File->throw( "not found: $file" );
    -r $file
      or WWW::Yahoo::Briefcase::Exception::File->throw( "unreadable: $file" );
    -f $file
      or WWW::Yahoo::Briefcase::Exception::File->throw( "not a file: $file" );
    my $size = -s $file;
    if ( $size > $self->maxsize ) {
	my $max = $self->maxsize;
	my $error = "file size ($size bytes) exceeds limit ($max bytes): $file";
	WWW::Yahoo::Briefcase::Exception::File->throw( $error );
    }

    # Get link for form to upload files.
    my $link_upload = $self->{__PACKAGE__.'_link_upload'}->{$folder};
    if ( not $link_upload ) {
	# Request form for selecting destination folder.
	$self->get( $self->{__PACKAGE__.'_link_addfiles'} );
	unless ( $self->content =~ m!Select${s}Folder! ) {
	    my $error = "unexpected response for add files form";
	    WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
	}

	# Select destination folder and submit.
	eval {
	    $self->field( '.sel' => $folder );
	    $self->click;
	};
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $@ ) if $@;
	unless ( $self->content =~ m!Upload${s}Files! ) {
	    my $error = "unexpected response for folder selection";
	    WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
	}

	# Get link for form to upload files.
	( $link_upload ) = grep { $_->name eq 'topf' } @{$self->links};
	unless ( defined $link_upload ) {
	    my $error = "no link to upload files";
	    WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
	}
	$link_upload = URI->new_abs( $link_upload->[0], $self->base );
	$self->{__PACKAGE__.'_link_upload'}->{$folder} = $link_upload;
    }

    # Get form to upload file.
    $self->get( $link_upload );
    unless ( $self->content =~ m!Browse${s}and${s}Upload! ) {
	my $error = "unexpected response for file upload form";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }

    # Upload file.
    my $field_set;
    eval {
	foreach my $input ( $self->current_form->inputs ) {
	    $input->type eq 'file' or next;
	    $input->file( undef );
	    next if $field_set;
	    $input->file( $file );
	    $input->filename( $filename );
	    ++$field_set;
	}
    };
    WWW::Yahoo::Briefcase::Exception::Interface->throw( $@ ) if $@;
    unless ( $field_set ) {
	my $error = "could not find field for file upload";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
    {
	{ package HTTP::Request::Common; our $DYNAMIC_FILE_UPLOAD }
	local $HTTP::Request::Common::DYNAMIC_FILE_UPLOAD = 1;
	local $SIG{__WARN__} =
	  sub { print STDERR $_[0] unless $_[0] =~ m!^<input> outside <form>! };
	$self->click;
    }
    unless ( $self->content =~ m!Confirm${s}Upload${s}Files! ) {
	$self->_check_for_error( 'upload file error' );
	my $error = "unexpected response for file upload";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
}

#-------------------------------------------------------------------------------
# Check current page for an error message (and throw exception if error).
# $self->_check_for_error( $error_context );
#-------------------------------------------------------------------------------
sub _check_for_error {
    my $self = shift;
    my $error = shift;

    return unless $self->content =~ m!Error${s}Message!;

    my( $error_code, $error_description );
    if ( $self->content =~ m!SetErrorMsg\("([^\"]*)"\)! ) {
	$error_code = $1;
    }
    elsif ( $self->content =~ m!scd\.yahoo\.com:(.*)! ) {
	$error_code = $1;
    }

    if ( $self->content =~ m!<b>Error</b><br>(.*)! ) {
	$error_description = $1;
	if ( defined $error_code and $error_code eq $error_description ) {
	    undef $error_code;
	}
    }
    elsif ( $self->content =~ m/Error${s}Message!(.*?)
				(?:<[^>]+href=|
				\n[a-z]+\d+.bc.scd.yahoo.com)/sx )
    {
	$error_description = $1;
	$error_description =~ s!<[^>]+>!!g;
	$error_description =~ s!${s}+! !g;
    }

    $error .= " ($error_code)" if defined $error_code;
    $error .= ": $error_description" if defined $error_description;
    WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
}

################################################################################

=pod

=head2 download

  $ybc->download( $folder, @files );
  $ybc->download( $folder, $file );
  $ybc->download( $folder, [ $file => $localdir ] );
  $ybc->download( $folder, [ $file => $localfile ], $file2, ... );

  $ybc->download( 'My Backups', 'dump.bz2' );
  $ybc->download( 'My Backups', [ 'dump.bz2' => '/var/backups' ] );
  $ybc->download( 'My Backups', [ '20030415.bz2' => '/var/backups/dump.bz2' ] );

Download specified file(s) from specified folder.  The full path name
of the existing folder on the remote system must be specified.
Optionally you can specify the name (and/or location) for the local
file.

Note that if the local file exists, the download will fail, unless you
have enabled the overwrite flag, in which case the file will be
overwritten.  If a local directory exists with the same name as the
file being downloaded, then the download will fail even if the
overwrite flag is set.

=cut

#-------------------------------------------------------------------------------
sub download {
    my $self = shift;
    if ( @_ < 2 ) {
	my $error = "folder and/or file(s) not specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    my $folder = $self->_canonical_folder( shift );

    # Get listing of objects in folder.
    $self->_list( $folder );

    # Make list of links for downloading files from folder.
    my $links = {};
    foreach my $link ( @{$self->links} ) {
	$link->url =~ m!http://[^/]+\.yahoofs\.com/[^?]+/([^/?]+)\?! or next;
	$links->{$1} = $link->url;
    }

    my @errors;
    foreach my $file ( @_ ) {
	my $location = ref $file ? $file->[1] : undef;
	$file = $file->[0] if ref $file;
	eval { $self->_download_file( $folder, $file, $location, $links ) };
	push( @errors, $@ ) if $@;
    }
    WWW::Yahoo::Briefcase::Exception::List->throw_list( @errors ) if @errors;
    (); # no return value
}

#-------------------------------------------------------------------------------
sub _download_file {
    my $self = shift;
    my( $folder, $file, $location, $links ) = @_;

    if ( $self->debug ) {
	my $to = defined $location ? " to $location" : '';
	warn "INFO: Downloading \"$file\" from \"$folder\"$to\n";
    }

    unless ( $links->{$file} ) {
	my $error = "file \"$file\" does not exist in \"$folder\"";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }

    # Determine location to place file.
    $location = $file unless defined $location;
    if ( -d $location ) {
	$location .= '/' unless substr( $location, -1 ) eq '/';
	$location .= $file;
    }
    if ( -d $location ) {
	my $error = "folder exists: $location";
	WWW::Yahoo::Briefcase::Exception::File->throw( $error );
    }
    if ( -e $location and not $self->overwrite ) {
	my $error = "file exists: $location";
	WWW::Yahoo::Briefcase::Exception::File->throw( $error );
    }
    $location =~ m!(.*)!; $location = $1; # untaint

    # Download file.
    $self->get( $links->{$file}, ':content_file' => $location );
    if ( $self->content ) {
	my $error = "unexpected response for file download";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
}

################################################################################

=pod

=head2 create

  $ybc->create( $folder, @newsubfolders );
  $ybc->create( $folder, $newsubfolder );

  $ybc->create( '/', 'Backups' );
  $ybc->create( 'My Photos', 'Vacation', 'Party' );

Create new subfolder(s) in an existing folder.  The full path of the
existing folder must be specified.

Note that if an existing object of the same name is found in the
folder, an error will be raised, unless you have enabled the overwrite
flag, in which case the object will be overwritten.  Note well that if
the pre-existing object to be overwritten is a folder then all its
contents are deleted!

=cut

#-------------------------------------------------------------------------------
sub create {
    my $self = shift;
    if ( @_ < 2 ) {
	my $error = "folder and/or subfolder(s) not specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    my $folder = $self->_canonical_folder( shift );

    # Get listing of objects in folder.
    my %exists = map { $_ => 1 } $self->list( $folder );

    # Find link for form to create folder.
    my $link = $self->find_link( text_regex => qr!Create${s}Folder! );
    unless ( defined $link ) {
	my $error = "no link to create folder";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
    $link = URI->new_abs( $link->[0], $self->base );

    # Check for overwrite flag and delete old subfolders.
    my $prefix = $self->overwrite;
    if ( $prefix ) {
	if ( $prefix eq '1' ) {
	    # Pre-delete subfolder(s).
	    my @delete = grep { $exists{$_} } @_;
	    $self->delete( $folder, @delete ) if @delete;
	    $prefix = '';
	}
	else {
	    # Post-delete subfolder(s), but delete old ones with prefix now.
	    my @delete = map { "$prefix$_" } grep { $exists{"$prefix$_"} } @_;
	    $self->delete( $folder, @delete ) if @delete;
	}
    }

    # Create subfolders.
    my( @rename, @errors );
    foreach my $subfolder ( @_ ) {
	unless ( defined $subfolder ) {
	    push( @errors, "subfolder not defined for \"$folder\"" );
	    next;
	}

	my $newfolder = $subfolder;
	if ( $prefix and $exists{$subfolder} ) {
	    # Create new subfolder with temporary name, which will be
	    # changed after the subfolder is created successfully and
	    # the old subfolder is deleted.
	    $newfolder = "$prefix$subfolder";
	}

	warn "INFO: Creating folder \"$newfolder\" in \"$folder\"\n"
	  if $self->debug;
	eval { $self->_create( $link, $newfolder ) };
	if ( $@ ) {
	    push( @errors, $@ );
	}
	elsif ( $subfolder ne $newfolder ) {
	    push( @rename, $subfolder );
	}
    }

    # Delete old subfolders and remove prefix from new subfolders.
    if ( @rename ) {
	my $overwrite = $self->overwrite( 0 ); # inhibit extraneous listings
	eval { $self->delete( $folder, @rename ) };
	push( @errors, $@ ) if $@;
	my %exists = map { $_ => 1 } $self->list( $folder );
	foreach my $subfolder ( @rename ) {
	    next if $exists{$subfolder};
	    eval {
		$self->rename( $folder, "$prefix$subfolder" => $subfolder )
	    };
	    push( @errors, $@ ) if $@;
	}
	$self->overwrite( $overwrite ); # restore previous value
    }

    WWW::Yahoo::Briefcase::Exception::List->throw_list( @errors ) if @errors;
    (); # no return value
}

#-------------------------------------------------------------------------------
sub _create {
    my $self = shift;
    my( $link, $subfolder ) = @_;

    # Get form to create folder.
    $self->get( $link );
    unless ( $self->content =~ m!Create${s}Folder! ) {
	my $error = "unexpected response for create folder form";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }

    # Create the folder.
    eval {
	$self->field( '.fnm' => $subfolder );
	$self->submit;
    };
    WWW::Yahoo::Briefcase::Exception::Interface->throw( $@ ) if $@;
    $self->_check_for_error( 'create folder error' );
}

################################################################################

=pod

=head2 rename

  $ybc->rename( $folder, $object => $newname );

  $ybc->rename( 'My Photos', 'pic.jpg' => 'picture' );

Renames an existing object.  The full path of the folder containing
the object must be specified.  Do not include the extension on the new
name.

If an object with the new name already exists, an error will occur,
unless you have enabled the overwrite flag, in which case the object
will be overwritten.  Note well that if the pre-existing object to be
overwritten is a folder then all its contents are deleted!

=cut

#-------------------------------------------------------------------------------
sub rename {
    my $self = shift;
    if ( @_ < 3 ) {
	my $error = "folder, object, and/or name not specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    if ( @_ > 3 ) {
	my $error = "extra arguments specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    my $folder = $self->_canonical_folder( shift );
    my( $object, $name ) = @_;
    defined $object
      or WWW::Yahoo::Briefcase::Exception::Usage->throw( "object not defined" );
    defined $name
      or WWW::Yahoo::Briefcase::Exception::Usage->throw( "name not defined" );

    if ( $self->overwrite ) {
	my( $type ) = $object =~ m!(\.[^.]*)$!;
	my $newname = defined $type ? "$name$type" : $name;
	if ( grep { $newname eq $_ } $self->list( $folder ) ) {
	    $self->delete( $folder, $newname );
	}
    }

    if ( $self->debug ) {
	my $folder = $folder; # do not modify original $folder
	$folder .= '/' unless $folder eq '/';
	warn "INFO: Renaming \"$folder$object\" to \"$name\"\n";
    }

    # Get listing of objects in folder.
    $self->_list( $folder );

    # Request form for editing the object.
    $self->_edit( $folder, $object );

    # Rename the object.
    my $form = $self->current_form;
    if ( $form->find_input( '.dnm' ) ) {
	# The object is a file.
	$self->field( '.dnm' => $name );
    }
    elsif ( $form->find_input( '.fnm' ) ) {
	# The object is a folder.
	$self->field( '.fnm' => $name );
    }
    else {
	my $error = "unexpected response for edit object";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
    $self->click;
    $self->_check_for_error( 'edit object error' );

    (); # no return value
}

#-------------------------------------------------------------------------------
# Request form to edit an object from a folder listing.  Unless the
# current page is a folder listing, call $self->_list before this method.
# $self->_edit( $folder, $object );
#-------------------------------------------------------------------------------
sub _edit {
    my $self = shift;
    my( $folder, $object ) = @_;

    $folder .= '/' unless substr( $folder, -1 ) eq '/';

    my $link_edit;
    foreach my $link ( @{$self->links} ) {
	$link->url =~ m!modify_briefcase.*?\.dir=([^&]+)! or next;
	my $dir = $self->_unescape( $1 );
	if ( $link->url =~ m!\.dnm=([^&]+)! ) {
	    # The object is a file.
	    my $dnm = $self->_unescape( $1 );
	    $object eq $dnm and $folder eq "$dir/" or next;
	    $link_edit = $link->url;
	    last;
	}
	else {
	    # The object is a folder.
	    $dir eq "$folder$object" or next;
	    $link_edit = $link->url;
	    last;
	}
    }
    unless ( $link_edit ) {
	my $error = "object not found: $folder$object";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }

    $self->get( $link_edit );
    unless ( $self->content =~ m!Modify${s}File|Edit${s}Folder${s}Settings! ) {
	my $error = "unexpected response for edit object";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
}

################################################################################

=pod

=head2 delete

  $ybc->delete( $folder, @objects );
  $ybc->delete( $folder, $objects ); # $objects is an array ref
  $ybc->delete( $folder, $object );

  $ybc->delete( 'My Photos', 'pic.jpg' );
  $ybc->delete( '/', 'backups' );

Delete the specified objects from the specified folder.  The full path
of the containing folder must be specified.  Note that for folder
objects, this method will recursively delete all subfolders and files
contained within it.

=cut

#-------------------------------------------------------------------------------
sub delete {
    my $self = shift;
    if ( @_ < 2 ) {
	my $error = "folder and/or object(s) not specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    $self->_perform_action( 'delete', @_ );
    (); # no return value
}

################################################################################

=pod

=head2 move

  $ybc->move( $folder => $destination, @objects );
  $ybc->move( $folder => $destination, $objects ); # $objects is an array ref
  $ybc->move( $folder => $destination, $object );

  $ybc->move( 'My Photos' => 'Old Photos', 'pic.jpg' );

Move objects to another folder.  The full path of the original folder
and the destination folder must be specified in $folder and
$destination, respectively.

Note that if an existing object of the same name is found in the
destination folder, it will automatically be renamed (for example,
'pic.jpg' might become 'pic---1.jpg'), unless you have enabled the
overwrite flag, in which case the object will be overwritten.  Note
well that if the pre-existing object to be overwritten is a folder
then all its contents are deleted!  Unlike upload and create, it is
not possible to use an overwrite-prefix, instead, the pre-existing
object is deleted before the move.

=cut

#-------------------------------------------------------------------------------
sub move {
    my $self = shift;
    if ( @_ < 3 ) {
	my $error = "folder, destination, and/or object(s) not specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    $self->_perform_action( 'move', @_ );
    (); # no return value
}

################################################################################

=pod

=head2 copy

  $ybc->copy( $folder => $destination, @objects );
  $ybc->copy( $folder => $destination, $objects ); # $objects is an array ref
  $ybc->copy( $folder => $destination, $object );

  $ybc->copy( 'My Photos' => 'Old Photos', 'pic.jpg' );

Copy objects to another folder.  The full path of the original folder
and the destination folder must be specified in $folder and
$destination, respectively.

Note that if an existing object of the same name is found in the
destination folder, it will automatically be renamed (for example,
'pic.jpg' might become 'pic---1.jpg'), unless you have enabled the
overwrite flag, in which case the object will be overwritten.  Note
well that if the pre-existing object to be overwritten is a folder
then all its contents are deleted!  Unlike upload and create, it is
not possible to use an overwrite-prefix, instead, the pre-existing
object is deleted before the copy.

=cut

#-------------------------------------------------------------------------------
sub copy {
    my $self = shift;
    if ( @_ < 3 ) {
	my $error = "folder, destination, and/or object(s) not specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    $self->_perform_action( 'copy', @_ );
    (); # no return value
}

#-------------------------------------------------------------------------------
# Perform specified action, which can be 'delete', 'move', or 'copy':
# $self->_perform_action( 'delete', $folder, @objects );
# $self->_perform_action( 'move', $folder, $destination, @objects );
# $self->_perform_action( 'copy', $folder, $destination, @objects );
#-------------------------------------------------------------------------------
sub _perform_action {
    my $self = shift;
    my $which = shift;

    my $folder = $self->_canonical_folder( shift );
    my $destination;
    unless ( $which eq 'delete' ) {
	$destination = $self->_canonical_folder( shift );
	if ( $self->overwrite ) {
	    my %exists = map { $_ => 1 } $self->list( $destination );
	    my @delete = grep { $exists{$_} } @_;
	    $self->delete( $destination, @delete ) if @delete;
	}
    }

    if ( $self->debug ) {
	( my $action = "\u$which" ) =~ s!e?$!ing!;
	my $to = $which eq 'delete' ? '' : " to \"$destination\"";
	foreach my $object ( @_ ) {
	    warn "INFO: $action \"$object\" from \"$folder\"$to\n";
	}
    }

    # Get listing of objects in folder.
    $self->_list( $folder );

    # Select objects to be deleted, moved, or copied.
    my( $selected, $unselected ) = $self->_select( @_ );
    my @errors;
    if ( @$unselected ) {
	my $error = "objects not found in \"$folder\": \"" .
	  join( '", "', @$unselected ) . '"';
	$error = WWW::Yahoo::Briefcase::Exception::Interface->new( $error );
	$error->throw unless @$selected;
	push( @errors, $error );
    }

    eval {
	# Request that the objects be deleted, moved, or copied and get
	# form for selecting destination folder.
	my $action = '.action' . ucfirst $which;
	$self->click( $action );

	if ( $which eq 'delete' ) {
	    # Confirm deletion.
	    $self->click( '.confdel' );
	}
	else {
	    # Select destination folder and submit.
	    $self->field( '.sel' => $destination );
	    $self->click( '.select' );
	}
    };
    push( @errors, WWW::Yahoo::Briefcase::Exception::Interface->new( $@ ) )
      if $@;
    eval { $self->_check_for_error( "$which objects error" ) };
    push( @errors, $@ ) if $@;

    WWW::Yahoo::Briefcase::Exception::List->throw_list( @errors ) if @errors;
}

#-------------------------------------------------------------------------------
# Select specified objects from a folder listing.  Unless the current
# page is a folder listing, call $self->_list before this method.
# my( $selected, $unselected ) = $self->_select( @objects );
#-------------------------------------------------------------------------------
sub _select {
    my $self = shift;

    my %unselected = map { $_ => 1 } @_;
    my %selected;

    eval {
	foreach my $input ( $self->current_form->inputs ) {
	    $input->type eq 'checkbox' or next;
	    $input->name eq '.docs' or next;
	    my( $value ) = $input->other_possible_values;
	    next unless $unselected{$value};
	    $input->value( $value );
	    $selected{$value} = 1;
	    delete $unselected{$value};
	}
    };
    WWW::Yahoo::Briefcase::Exception::Interface->throw( $@ ) if $@;

    ( [ keys %selected ], [ keys %unselected ] );
}

################################################################################

=pod

=head2 list

  my @contents = $ybc->list( $folder );
  my @contents = $ybc->list;
  my @contents = $ybc->list( '' );

  my @photos = $ybc->list( 'My Photos' );
  my @vacation_photos = $ybc->list( 'My Photos/Vacation' );

List contents of an existing folder.  The full path of the existing
folder must be specified.  Specify '', '/' or undef for a top-level
listing.

The returned list contains objects from the class
L<WWW::Yahoo::Briefcase::Object|WWW::Yahoo::Briefcase::Object>.  You
can access details of each object as follows:

  foreach my $object ( @contents ) {
      my $location = $object->location; # parent folder
      my $name = $object->name;
      my $fullname = $object->fullname;
      my $fullname = "$object"; # stringification returns fullname of object
      my $type = $object->type;
      my $size = $object->size; # in bytes
      my $time = $object->time; # last-modified-time (seconds since epoch)
  }

Note that the size is rounded off to the nearest kilobyte (it is
reported that way in the folder listing) and the time is rounded off
to midnight GMT (only the date is reported in the folder listing).

=cut

#-------------------------------------------------------------------------------
my %size_factor = ( KB => 1024, MB => 1_048_576, item => 1, items => 1 );
my %month;
@month{qw( Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec )} = ( 0 .. 11 );
sub list {
    my $self = shift;
    if ( @_ < 1 ) {
	my $error = "folder not specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    if ( @_ > 1 ) {
	my $error = "extra arguments specified";
	WWW::Yahoo::Briefcase::Exception::Usage->throw( $error );
    }
    my $folder = $self->_canonical_folder( shift );

    my $cache = $self->{__PACKAGE__.'_list_cache'};
    if ( $cache and $cache->{folder} eq $folder
	 and time - $cache->{time} <= $list_cache_time )
    {
	# Note: there will be no $cache->{contents} if the "_list"
	# method was recently called but the response was not parsed.
	# In this case we don't need to call the "_list" method again,
	# but we do need to parse the response.
	warn "INFO: Listing \"$folder\" (from cache)\n" if $self->debug;
	return @{$cache->{contents}} if $cache->{contents};
    }
    else {
	warn "INFO: Listing \"$folder\"\n" if $self->debug;
	$self->_list( $folder );
    }

# Old way, which is more reliable, but doesn't get any details:
#    my @contents;
#    foreach my $input ( $self->current_form->inputs ) {
#	$input->type eq 'checkbox' or next;
#	$input->name eq '.docs' or next;
#	my( $value ) = $input->other_possible_values;
#	push( @contents, $value );
#    }
#    @contents;

    my $html = $self->response->content;
    return if $html =~ m!this folder is currently empty!i;

    my @contents;
    my @headers = map { "\\A$s*(?:<[^>]*>$s*)*$_(?:$s*<[^>]*>)*$s*\\Z" }
      ( 'Name', 'Type', 'Size', 'Last Modified' );
    my $te = HTML::TableExtract->new( headers => \@headers, keep_html => 1 );
    $te->parse( $html );
    my @ts = $te->table_states;
    if ( @ts < 1 ) {
	my $error = "no table listing found";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
    if ( @ts > 1 ) {
	my $error = "more than one table listing found";
	WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
    }
    foreach my $row ( $ts[0]->rows ) {
	# We have to extract the filename from the link because if it
	# is too long it will be truncated in the text.
	unless ( $row->[0] =~ m!\.dir=([^&]+)! # folder
		 or $row->[0] =~ m!<a href="[^"]+/([^?"]+)! ) # file
	{
	    my $error = "could not find link in name field: $row->[0]";
	    WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
	}
	my $name = $1;

	foreach my $cell ( @$row ) {
	    $cell =~ s!<[^>]*>!!g;
	    $cell =~ s!^$s*!!;
	    $cell =~ s!$s*$!!;
	}

	my( $truncated_name, $type, $size, $time ) = @$row;
	unless ( $type eq 'folder' ) {
	    substr( $name, - length( $type ) ) = ''; # remove extension
	}

	if ( $size =~ m!^(\d+) (\w+)$! and defined $size_factor{$2} ) {
	    $size = $1 * $size_factor{$2};
	}
	else {
	    my $error = "unrecognized size for $name in listing: $size";
	    WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
	}

	if ( $time =~ m!^(\d\d?)-([A-Z][a-z][a-z])-(\d\d\d\d)$!
	     and defined $month{$2} )
	{
	    $time = Time::Local::timegm( 0, 0, 0, $1, $month{$2}, $3-1900 );
	}
	else {
	    my $error = "unrecognized time for $name in listing: $time";
	    WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
	}

	my @args = ( location => $folder, name => $name, type => $type,
		     size => $size, time => $time );
	push( @contents, WWW::Yahoo::Briefcase::Object->new( @args ) );
    }
    $self->{__PACKAGE__.'_list_cache'}->{contents} = \@contents;
    @contents;
}

#-------------------------------------------------------------------------------
sub _list {
    my $self = shift;
    my $folder = shift;

    my $cache = $self->{__PACKAGE__.'_list_cache'};
    return if $cache and $cache->{folder} eq $folder
      and time - $cache->{time} <= $list_cache_time;

    my $link_folder = $self->{__PACKAGE__.'_link_folder'}->{$folder};

    if ( not $link_folder ) {
	$self->get( $self->{__PACKAGE__.'_link_myfolders'} );
	foreach my $link ( @{$self->links} ) {
	    $link->url =~ m!\.dir=([^&]+)! or next;
	    my $this_folder = $self->_unescape( $1 );
	    next if defined $self->{__PACKAGE__.'_link_folder'}->{$this_folder};
	    my $url = $link->url . '&.begin=9999'; # to Show All
	    $self->{__PACKAGE__.'_link_folder'}->{$this_folder} = $url;
	    if ( $folder eq $this_folder ) {
		$link_folder = $url;
	    }
	}
	unless ( $link_folder ) {
	    my $error = "folder not found: $folder";
	    WWW::Yahoo::Briefcase::Exception::Interface->throw( $error );
	}
    }

    $self->get( $link_folder );
    $self->_check_for_error( 'folder listing error' );

    # Prepare cache of this listing.
    $self->{__PACKAGE__.'_list_cache'} = { folder => $folder, time => time };
}

#-------------------------------------------------------------------------------
# $self->_canonical_folder( $folder );
#-------------------------------------------------------------------------------
sub _canonical_folder {
    my $self = shift;
    my $folder = shift;
    $folder = '/' unless defined $folder;
    $folder = "/$folder" unless substr( $folder, 0, 1 ) eq '/';
    $folder =~ s!/$!! if substr( $folder, -1 ) eq '/' and $folder ne '/';
    $folder;
}

#-------------------------------------------------------------------------------
sub _unescape {
    my $self = shift;
    local( $_ ) = shift;
    tr!+! !;
    s!%([0-9A-Fa-f]{2})!chr(hex($1))!eg;
    $_;
}

################################################################################

=pod

=head1 SEE ALSO

L<http://briefcase.yahoo.com/>

=head1 AUTHOR

Ken Neighbors, Ph.D.
ken@nsds.com

=head1 COPYRIGHT AND LICENSE

Copyright 2003 by Ken Neighbors

This module is free software; you can redistribute it and/or modify it
under the same terms as Perl itself, either Perl version 5.000 or (at
your option) any later version of Perl.  That is, you can redistribute
this module and/or modify it under the terms of either:

  a) the GNU General Public License as published by the Free Software
     Foundation; either version 1, or (at your option) any later version,
     or

  b) the "Artistic License" which comes with Perl, or

  c) the license terms of Perl versions later than 5.000

=cut

1;
