#!/usr/bin/perl -Tw
################################################################################
package WWW::Yahoo::Briefcase::Exceptions;
# $Id: Exceptions.pm,v 1.3 2003/12/05 00:03:19 ken Exp $

=pod

=head1 NAME

WWW::Yahoo::Briefcase::Exceptions - exceptions related to briefcase access

=head1 SYNOPSIS

  require WWW::Yahoo::Briefcase;
  eval { something };
  if ( ref $@ ) {
      if ( $@->isa( 'WWW::Yahoo::Briefcase::Exception::Usage' ) ) {
	  die "Usage error: $@\n";
      }
      elsif ( $@->isa( 'WWW::Yahoo::Briefcase::Exception::Signin' ) ) {
	  die "Signin error: $@\n";
      }
      elsif ( $@->isa( 'WWW::Yahoo::Briefcase::Exception::Server' ) ) {
	  die "Server error: $@\n";
      }
      elsif ( $@->isa( 'WWW::Yahoo::Briefcase::Exception::Interface' ) ) {
	  die "Interface error: $@\n";
      }
      elsif ( $@->isa( 'WWW::Yahoo::Briefcase::Exception::File' ) ) {
	  die "File error: $@\n";
      }
      elsif ( $@->isa( 'WWW::Yahoo::Briefcase::Exception::List' ) ) {
	  die "Multiple errors: $@\n";
      }
  }
  die $@ if $@;

=head1 DESCRIPTION

This module defines several classes of exceptions as described below.

=cut

################################################################################
use 5.006;
use strict;
use warnings;
our $VERSION = '0.81';
use Exception::Class
  ( 'WWW::Yahoo::Briefcase::Exception' =>
    { description => 'briefcase exception' } );

################################################################################

=pod

=head2 WWW::Yahoo::Briefcase::Exception::Usage

Incorrect or invalid usage of method parameters.  Note that a full
stack trace will be printed when this exception is stringified to help
you find out where the mistake occurred.  To disable the stack trace,
include this line somewhere after "require WWW::Yahoo::Briefcase":

  WWW::Yahoo::Briefcase::Exception::Usage->Trace(0);

=cut

#-------------------------------------------------------------------------------
use Exception::Class
  ( 'WWW::Yahoo::Briefcase::Exception::Usage' =>
    { isa => 'WWW::Yahoo::Briefcase::Exception',
      description => 'usage error' } );

WWW::Yahoo::Briefcase::Exception::Usage->Trace(1);

################################################################################

=pod

=head2 WWW::Yahoo::Briefcase::Exception::Signin

Invalid userid, password, or cookie.

=cut

#-------------------------------------------------------------------------------
use Exception::Class
  ( 'WWW::Yahoo::Briefcase::Exception::Signin' =>
    { isa => 'WWW::Yahoo::Briefcase::Exception',
      description => 'signin error' } );

################################################################################

=pod

=head2 WWW::Yahoo::Briefcase::Exception::Server

Yahoo returned a server error for an HTTP request.  The request and
response are available as fields of this exception, for example,

  eval { something };
  if ( $@ and $@->isa( 'WWW::Yahoo::Briefcase::Exception::Server' ) ) {
      my $request_line = $@->request->method . ' ' . $@->request->uri;
      my $status_line = $@->response->status_line;
  }

=cut

#-------------------------------------------------------------------------------
use Exception::Class
  ( 'WWW::Yahoo::Briefcase::Exception::Server' =>
    { isa => 'WWW::Yahoo::Briefcase::Exception',
      description => 'server error',
      fields => [ 'request', 'response' ] } );

sub WWW::Yahoo::Briefcase::Exception::Server::full_message {
    my $self = shift;
    $self->message .
      '(' . $self->request->method . ' ' . $self->request->uri .
      ' => ' . $self->response->status_line . ')';
}

################################################################################

=pod

=head2 WWW::Yahoo::Briefcase::Exception::Interface

Yahoo returned an unexpected response to an HTTP request.

=cut

#-------------------------------------------------------------------------------
use Exception::Class
  ( 'WWW::Yahoo::Briefcase::Exception::Interface' =>
    { isa => 'WWW::Yahoo::Briefcase::Exception',
      description => 'interface error' } );

################################################################################

=pod

=head2 WWW::Yahoo::Briefcase::Exception::File

Problem accessing a local file.

=cut

#-------------------------------------------------------------------------------
use Exception::Class
  ( 'WWW::Yahoo::Briefcase::Exception::File' =>
    { isa => 'WWW::Yahoo::Briefcase::Exception',
      description => 'file error' } );

################################################################################

=pod

=head2 WWW::Yahoo::Briefcase::Exception::List

Holds a list of errors, which can be accessed via the errors method:

  eval { something };
  if ( $@ and $@->isa( 'WWW::Yahoo::Briefcase::Exception::List' ) ) {
      foreach my $error ( $@->errors ) {
	  warn $error->message . "\n" . $error->trace->as_string . "\n";
	  warn "------------------------------------\n";
      }
      exit;
  }

=cut

#-------------------------------------------------------------------------------
use Exception::Class
  ( 'WWW::Yahoo::Briefcase::Exception::List' =>
    { isa => 'WWW::Yahoo::Briefcase::Exception',
      description => 'list of errors',
      fields => [ 'errors' ] } );

sub WWW::Yahoo::Briefcase::Exception::List::throw_list {
    my $class = shift;
    return unless @_;
    $_[0]->rethrow if @_ == 1;
    my @errors;
    foreach my $error ( @_ ) {
	if ( $error->isa( 'WWW::Yahoo::Briefcase::Exception::List' ) ) {
	    push( @errors, @{ $error->errors } );
	}
	elsif ( ref $error eq 'ARRAY' ) {
	    push( @errors, @$error );
	}
	else {
	    push( @errors, $error );
	}
    }
    my $error = join( ";\n", map { $_->error } @errors ) . "\n";
    $class->throw( error => $error, errors => \@errors );
}

################################################################################

=pod

=head1 SEE ALSO

L<WWW::Yahoo::Briefcase|WWW::Yahoo::Briefcase>,
L<WWW::Yahoo::Briefcase::File|WWW::Yahoo::Briefcase::File>,
L<Exception::Class|Exception::Class>

=head1 AUTHOR

Ken Neighbors, Ph.D.
ken@nsds.com

=head1 COPYRIGHT AND LICENSE

Copyright 2003 by Ken Neighbors

This module is free software; you can redistribute it and/or modify it
under the same terms as Perl itself, either Perl version 5.000 or (at
your option) any later version of Perl.  That is, you can redistribute
this module and/or modify it under the terms of either:

  a) the GNU General Public License as published by the Free Software
     Foundation; either version 1, or (at your option) any later version,
     or

  b) the "Artistic License" which comes with Perl, or

  c) the license terms of Perl versions later than 5.000

=cut

1;
